/********************************************/
/*********** RE52 - Projet Dicom ************/
/********************************************/
/*** BOURGON Jean-Xavier / LOURD Rodolphe ***/
/********************************************/


// ENVOI DU MESSAGE (AVERTISSEMENT CLIENT DESTINATAIRE EXISTANT/INEXISTANT) A REALISER
// FAIRE EN SORTE DE POUVOIR IDENTIFIER PLUSIEURS CLIENTS DIFFUSEURS
// IMPORTANT => PROBLEME PRINCIPAL DE DESCRIPTEUR = VOIR COMMENT UN CLIENT CLASSIQUE EN OBTIENT UN ET FAIRE PAREIL POUR LE DIFFUSEUR


#include "message.h"


/*** dclarations du serveur ***/

int sock_rm,sock_dc,acc_rm;  // descripteurs de sockets
int reuse=1;                 // port rutilisable
t_message message;
int nbclients=0;             // indique le nombre de clients connects
char *thosts[64];            // tableau contenant les noms des clients
char nom_dclient[20];        // nom de l'metteur
char nom_dest[20];           // nom du destinataire
int test_dest;               // test de prsence du destinataire
char mess_diff[20];          // message  diffuser

struct sockaddr_in serveur_rm;  // socket de rception des messages
struct sockaddr_in serveur_dc;  // socket de diffusion des messages
struct hostent *hp;

extern void emettre_message(int client);
int recevoir_message(int client);
void diffuser(t_message msg);
void fermer_connexions();
void attente_diffuseur();
void traitant_FIN();

/*** dclarations de l'metteur ***/

int sock;       // descripteur de socket
char nom[20];   // nom de l'metteur

void recevoir_message_dif(int client);
void traitant_SIGPIPE(int n);
void traitant_SIGINT(int n);

/*** dclarations du rcepteur ***/

int at;            // descripteur de socket
char nom_cli[20];  // nom du rcepteur

extern void emettre_message(int client);
void recevoir_message_cli(int client);
void traitant_SIGNAL(int n);


/*** fonctions du serveur ***/

/* diffuser : fonction permettant de diffuser un message  tous les rcepteurs connects */
void diffuser(t_message msg)
{
  int i;
  for (i=0;i<32;i++)
  {
    if (thosts[i]!=NULL)
    {
      emettre_message(i);
    }
  }
}

/* fermer_connexions : fonction permettant de fermer toutes les connexions avec les rcepteurs */
void fermer_connexions()
{
  int i,c;
  for (i=0;i<32;i++)
  {
    if (thosts[i]!=NULL)
    {
      close(i);
      free(thosts[i]);
      thosts[i]=NULL;
    }
  }
  nbclients=0;
}

/* attente_diffuseur : fonction permettant de mettre le serveur en attente d'un metteur */
void attente_diffuseur()
{
  int lg;
  printf("\nEmetteur non connect...\n");
  lg=sizeof(serveur_rm);
  acc_rm=accept(sock_rm,(struct sockaddr *)&serveur_rm,(int *)&lg);
  if (acc_rm==-1)
  {
    perror("Erreur accept : ");
    exit(1);
  }
}

/* traitant_FIN : fonction permettant d'arrter le serveur proprement */
void traitant_FIN(int n)
{
  message.type=D_SERVEUR;
  diffuser(message);
  close(sock_dc);
  close(sock_rm);
  close(acc_rm);
  printf("\nA bientt...\n");
  exit(1);
}

/* recevoir_message : fonction permettant de recevoir un message d'un metteur/rcepteur */
int recevoir_message(int client)
{
  int test;
  test=read(client,&message,sizeof(message));
  if (test==-1)
  {
    perror("Erreur read : ");
    exit(1);
  }
  return (int)test;
}

/*** fonctions de l'metteur ***/

/* recevoir_message_dif : fonction permettant de recevoir un message du serveur */
void recevoir_message_dif(int client)
{
  int test;
  test=read(client,&message,sizeof(message));
  if (test==-1)
  {
    perror("Erreur read : ");
    exit(1);
  }
  if (test==0)
  {
    printf("\nConnexion avec le serveur coupe.\n");
    close(sock);
    exit(1);
  }
}

/* traitant_SIGPIPE : fonction appele par SIGPIPE permettant de dtecter la dconnexion brutale du serveur */
void traitant_SIGPIPE(int n)
{
  printf("\nConnexion avec le serveur perdue...\n");
  close(sock);
  raise(SIGPIPE);
  exit(1);
}

/* traitant_SIGINT : fonction appele par SIGINT permettant de raliser la dconnexion de l'metteur auprs du serveur */
void traitant_SIGINT(int n)
{
  message.type=D_DCLIENT;
  strcpy(message.mess,"");
  emettre_message(sock);
  close(sock);
  printf("\nDconnexion...\n");
  exit(1);
}

/*** fonctions du rcepteur ***/

/* recevoir_message_cli : fonction permettant de recevoir un message du serveur */
void recevoir_message_cli(int client)
{
  int test;
  test=read(client,&message,sizeof(message));
  if (test==-1)
  {
    perror("Erreur read : ");
    exit(1);
  }
  if (test==0)
  {
    printf("\nConnexion avec le serveur coupe.\n");
    close(sock);
    exit(1);
  }
}

/* traitant_SIGNAL : fonction appele par SIGINT permettant de raliser la dconnexion du rcepteur auprs du serveur */
void traitant_SIGNAL(int n)
{
  message.type=D_CLIENT;
  emettre_message(sock);
  close(sock);
  printf("\nDconnexion...\n");
  exit(1);
}


/***************************/
/*** Programme principal ***/
/***************************/

void main()
{
  
  /* variables du serveur */

  int lg,c,test,acc_dc,retval,i;
  struct timeval tv;
  char nom[20];  // nom du serveur
  fd_set fdmask;
  struct sigaction action;

  /* variables de l'metteur */

  char hote[50];  // nom de l'hte
  struct sockaddr_in serveur;
  struct hostent *hp;

  /* variables globales */

  char choix;
  choix='0';
  
  /* boucle de choix du poste utilis */

  while ((choix!='1')&&(choix!='2')&&(choix!='3'))
  {
    system("clear");
    printf("\n* Bienvenue sur l'application Dicom - par J.X. Bourgon et R. LOURD *\n");
    printf("\n* taper 1 pour tre serveur");
    printf("\n* taper 2 pour tre metteur");
    printf("\n* taper 3 pour tre rcepteur\n");
    scanf("%c",&choix);
  }

  switch (choix)
  {
    case '1':   // choix du serveur

      /* installation d'un handler pour le signal SIGINT */
      signal(SIGINT,traitant_FIN);
      gethostname(nom,sizeof(nom));
      printf("Serveur dmarr ; le nom de la machine est : %s\n",nom);
      hp=gethostbyname(nom);

      /* cre la socket d'coute pour communiquer avec les rcepteurs */
      sock_dc=socket(AF_INET,SOCK_STREAM,0);      // socket de demande de connexion des rcepteurs
      if (sock_dc==-1)
      {
        perror("Erreur socket_dc : ");
        exit(1);
      }

      /* cre la socket d'coute pour communiquer avec l'metteur */
      sock_rm=socket(AF_INET,SOCK_STREAM,0);      // socket de rception des messages */
      if (sock_rm==-1)
      {
        perror("Erreur socket_rm : ");
        exit(1);
      }

      /* initialisation de la structure serveur_rm ct metteur */
      bzero(&serveur_rm,sizeof(serveur_rm));
      serveur_rm.sin_family=AF_INET;
      serveur_rm.sin_port=htons(PORT_RM);
      bcopy(hp->h_addr,(char *)&serveur_rm.sin_addr,hp->h_length);

      /* initialisation de la structure serveur_dc ct rcepteur */
      bzero(&serveur_dc,sizeof(serveur_dc));
      serveur_dc.sin_family=AF_INET;
      serveur_dc.sin_port=htons(PORT_DC);

      /* attachement de sock_rm */
      setsockopt(sock_rm,SOL_SOCKET,SO_REUSEADDR,(int *)&reuse,sizeof(reuse));
      if (bind(sock_rm,(struct sockaddr *)&serveur_rm,sizeof(serveur_rm))==-1)
      {
        perror("Erreur bind_rm : ");
        exit(1);
      }

      /* attachement de sock_dc */
      setsockopt(sock_dc,SOL_SOCKET,SO_REUSEADDR,(int *)&reuse,sizeof(reuse));
      if (bind(sock_dc,(struct sockaddr *)&serveur_dc,sizeof(serveur_dc))==-1)
      {
        perror("Erreur bind_dc : ");
        exit(1);
      }

      /* cration de la file d'attente de la demande de connexion de l'metteur */
      if (listen(sock_rm,1)==-1)
      {
        perror("Erreur listen_rm : ");
        exit(1);
      }

      /* cration de la file d'attente des demandes de connexion des rcepteurs */
      if (listen(sock_dc,5)==-1)
      {
        perror("Erreur listen_dc : ");
        exit(1);
      }

      /* attente de la connexion d'un metteur */
      attente_diffuseur();
      printf("\nConnexion d'un metteur en cours ; attente de son nom...\n");

      /* boucle du serveur */
      while (1)
      {
        FD_ZERO(&fdmask);  // effacer tous les descripteurs associs a fdmask
        FD_SET(sock_dc,&fdmask);      // ajout du descripteur sock_dc
        FD_SET(acc_rm,&fdmask);       // ajout du descripteur sock_rm

        /* ajout des descripteurs de tous les rcepteurs connects */
        for (i=0;i<33;i++)
        {
          if (thosts[i]!=NULL)
          {
            FD_SET(i,&fdmask);
          }
        }

        /* coute de tous les descripteurs de sockets pendant 15 secondes maxi */
        tv.tv_sec=15;
        tv.tv_usec=0;
        retval=select(32,&fdmask,NULL,NULL,&tv);
        if (retval<0)
        {
          printf("Erreur select number %i: \n",errno);
          perror("Erreur select ");
          exit(1);
        }

        /* si rception de l'metteur */
        if (FD_ISSET(acc_rm,&fdmask))
        {
          test=recevoir_message(acc_rm);

          /* test si rception d'informations de l'metteur */
          if (message.type==INFO_DCLIENT)
          {
            printf("\nConnexion de l'metteur %s effectue...\n",message.mess);
            strcpy(nom_dclient,message.mess);

            /* rcupration du nom de l'metteur qui vient de se connecter */
            printf("\nnom_dclient : %s\n",nom_dclient);
            thosts[32]=(char *)malloc(strlen(message.mess)*sizeof(char));
            strcpy(thosts[32],message.mess);
            //printf("\nValeur de thosts[32] : %s\n",thosts[32]);
            test=recevoir_message(acc_rm);
            strcpy(mess_diff, message.mess);
          }

          /* informations sur le destinataire du message */
          if (message.type==DEST_DCLIENT)
          {
            printf("\nLe destinataire du message est %s\n",message.mess);
            strcpy(nom_dest,message.mess);
            test_dest=0;
            for (i=0;i<32;i++)
            {
              if (thosts[i]!=NULL)
              {
                //printf ("\nDEBUT FOR\n");
                if (strcmp(thosts[i],nom_dest)==0)
                {
                  //printf ("\nDEBUT IF\n");
                  test_dest=1;
                }
              }
            }
            if (test_dest==1)
            {
              //printf ("\nDEBUT IF2\n");
              printf("\nLe rcepteur destinataire %s est connect\n",nom_dest);
            }
            else
            {
              //printf ("\nDEBUT ELSE2\n");
              printf("\nLe rcepteur destinataire %s n'est pas connect\n",nom_dest);

              message.type=BRABRABRA;
              strcpy(message.mess,"brabrabra");
              diffuser(message);
              //printf("\nNotification envoye au diffuseur\n");
            }
            test=recevoir_message(acc_rm);
            strcpy(mess_diff, message.mess);
          }

          /* test si le message doit tre envoy  tous les rcepteurs */
          if (message.type==M_DCLIENT)
          {
            printf("Message envoy : %s\n",message.mess);
            diffuser(message);
            strcpy(mess_diff,message.mess);
          }

          /* test si l'metteur envoie une demande de dconnexion */
          if (message.type==D_DCLIENT || test==0)
          {
            printf("L'metteur %s vient de se dconnecter...\n",nom_dclient);
            message.type=D_DCLIENT;
            fermer_connexions();
            close(acc_rm);
            attente_diffuseur();
          }
        }

        /* test si un rcepteur envoie une demande de dconnexion (socket d'coute) */
        if (FD_ISSET(sock_dc,&fdmask))
        {
          /* cration d'un descripteur de socket pour les rcepteurs qui demandent une connexion puis tablissement de celle-ci */
          printf("\nAttente de connexion d'un rcepteur...\n");
          lg=sizeof(serveur_dc);
          acc_dc=accept(sock_dc,(struct sockaddr *)&serveur_dc,(int *)&lg);
          if (acc_dc==-1)
          {
            perror ("Erreur accept sur sock_dc : ");
            exit(1);
          }
          test=recevoir_message(acc_dc);
          if (message.type==C_CLIENT)
          {
            printf("Le message  recevoir est : %s\n",mess_diff);
            thosts[acc_dc]=(char *)malloc(strlen(message.mess)*sizeof(char));
            strcpy(thosts[acc_dc],message.mess);
            nbclients++;
            printf("Le rcepteur %s est connect...\n", thosts[acc_dc]);
            printf("Il y a %d rcepteur(s) connects...\n\n",nbclients);
            message.type=INFO_SERVEUR;
            strcpy(message.mess,nom);
            emettre_message(acc_dc);
            message.type=INFO_DCLIENT;
            strcpy(message.mess,nom_dclient);
            emettre_message(acc_dc);
            message.type=M_DCLIENT;
            strcpy(message.mess,mess_diff);
            emettre_message(acc_dc);
          }
        }

        /* test si on reoit la dconnexion d'un rcepteur */
        for (i=0;i<32;i++)
        {
          if (thosts[i]!=NULL)
          {
            if (FD_ISSET(i,&fdmask))
            {
              test=recevoir_message(i);
              if (message.type==D_CLIENT || test==0)
              {
                printf("\nLe rcepteur %s est dconnect.\n",thosts[i]);
                thosts[i]=NULL;
                close(i);
                nbclients--;
                printf("\nIl y a %d rcepteurs connects.\n",nbclients);
                free(thosts[i]);
              }
            }
          }
        }
      }

      break;

    case '2': // choix de l'metteur

      printf("\nVous tes metteur ; veuillez indiquer le nom d'hte ou l'adresse IP du serveur...\n");
      scanf("%s",&hote);

      /* installation d'un handler pour le signal SIGINT */
      signal(SIGINT,traitant_SIGINT);

      /* installation d'un handler pour le signal SIGPIPE */
      signal(SIGPIPE,traitant_SIGPIPE);
      hp=gethostbyname(hote);

      /* cration de la socket de communication avec le serveur */
      sock=socket(AF_INET,SOCK_STREAM,0);
      if (sock==-1)
      {
        perror("Erreur socket : ");
        exit(1);
      }

      /* initialisation de la structure serveur */
      bzero(&serveur,sizeof(serveur));
      serveur.sin_family=AF_INET;
      serveur.sin_port=htons(PORT_RM);
      bcopy(hp->h_addr,(char *)&serveur.sin_addr,hp->h_length);

      /* demande de connexion sur le serveur */
      printf("\nConnexion sur le serveur de diffusion %s...\n",hp->h_name);
      if (connect(sock,(struct sockaddr *)&serveur,sizeof(serveur))==-1)
      {
        perror("Erreur connect ");
        exit(1);
      }

      /* envoie le nom de l'metteur au serveur */
      message.type=INFO_DCLIENT;
      printf("Entrez votre identit :\n");
      scanf("%s",&nom);
      strcpy(message.mess,nom);
      printf("le nom du client_diff est %s\n",message.mess);
      emettre_message(sock);
        
      /* envoi du destinataire auquel diffuser le message */
      message.type=DEST_DCLIENT;
      printf("Nom du destinataire : \n");
      scanf("%s",&nom);
      strcpy(message.mess,nom);
      //printf("le nom du destinataire est %s\n",message.mess);
      emettre_message(sock);
        
      /* boucle de traitement des messages envoys au serveur */
      while (1)
      {
        message.type=M_DCLIENT;
        printf("Message  envoyer : ");
        gets(message.mess);
        emettre_message(sock);

        /* messages reus du serveur */
        if (message.type==BRABRABRA)
        printf("\nDestinataire inconnu\n");
      }

      break;
	
    case '3': // choix du rcepteur

      printf("\nVous tes rcepteur ; veuillez indiquer le nom d'hte ou l'adresse IP du serveur...\n");
      scanf("%s",&hote);

      /* installation d'un handler pour le signal SIGINT */
      signal(SIGINT,traitant_SIGNAL);
      hp=gethostbyname(hote);
      
      /* cration de la socket de communication avec le serveur */
      sock=socket(AF_INET,SOCK_STREAM,0);
      if (sock==-1)
      {
        perror("Erreur socket : ");
        exit(1);
      }

      /* initialisation de la structure serveur */
      bzero(&serveur,sizeof(serveur));
      serveur.sin_family=AF_INET;
      serveur.sin_port=htons(PORT_DC);
      bcopy(hp->h_addr,(char *)&serveur.sin_addr,hp->h_length);

      /* demande de connexion sur le serveur */
      printf("\nDemande de connexion ...\n");
      if (connect(sock,(struct sockaddr *)&serveur,sizeof(serveur))==-1)
      {
        perror("Erreur connect ");
        exit(1);
      }

      /* envoie le nom du rcepteur au serveur */
      message.type=C_CLIENT;
      printf("Entrez votre identit :\n");
      scanf("%s",nom_cli);
      strcpy(message.mess,nom_cli);
      emettre_message(sock);

      /* boucle de traitement des messages arrivant du serveur */
      while (1)
      {
        /* lecture du message */
        recevoir_message_cli(sock);

        /* rception d'informations concernant le serveur */
        if (message.type==INFO_SERVEUR)
        {
          printf("\nBienvenue sur le serveur %s\n",message.mess);
        }

        /* rception d'informations concernant l'metteur */
        if (message.type==INFO_DCLIENT)
        {
          printf("\nL'metteur est %s\n",message.mess);
        }

        /* si le message vient de l'metteur alors on l'affiche */
        if (message.type==M_DCLIENT)
        {
          printf("\nLe message est %s\n",message.mess);
        }

        /* si le serveur se dconnecte alors on dconnecte le rcepteur */
        if (message.type==D_SERVEUR)
        {
          printf("\nConnexion coupe par le serveur...\n");
          close(sock);
          exit(1);
        }
        
        /* si l'metteur se dconnecte alors on dconnecte le rcepteur */
        if (message.type==D_DCLIENT)
        {
          printf("\nLe client diffuseur vient de se deconnecter...\n\n");
          close(sock);
          exit(1);
        }
      }

      break;

  }
}

